/*
   This program parses and then evaluates strings from Language_7.
*/

public class Language_7_Examples
{
   public static void main(String[] args)
   {
      String[] programs =
      {
         "( prog " +
           "( var y 6 ) " +
           "( fun f ( lambda x ( * x y ) ) ) " +  // y is a non-local reference
           "( apply f 5 ) )",

         "( prog " +
           "( fun f ( lambda x ( * x x ) ) ) " +
           "( fun g ( lambda x ( + x ( apply f x ) ) ) ) " +
           "( apply g 5 ) )",

         "( prog " +
           "( var w 10 ) " +
           "( fun f ( lambda x ( * x x ) ) ) " +
           "( var z ( begin " +
                     "( var x 0 ) " +
                     "( begin ( var y 2 ) " +
                             "( begin " +
                               "( set x w ) " +
                               "( apply f ( + x y ) ) ) ) ) ) " +
           "z )",

         "( prog " +
           "( var w 10 ) " +
           "( fun f ( lambda x y z ( + x y z w ) ) ) " + // w is a non-local reference
           "( var x ( apply f 1 2 3 ) ) " +
           "( var y ( apply f 4 5 x ) ) " +
           "( var z ( apply f ( + w x y ) 10 20 ) ) )",

         "( prog " +
           "( var n 1 ) " +
           "( fun f ( lambda x ( * n x ) ) ) " +
           "( fun g ( lambda x ( begin " +
                                "( var n 5 ) " +
                                "( apply f x ) ) ) ) " +
           "( apply g 2 ) )",  // result =  2 with static scope
                               // result = 10 with dynamic scope

         "( prog " +
           "( var n 1 ) " +
           "( fun f ( lambda x ( * n x ) ) ) " +
           "( fun g ( lambda x ( begin " +
                                "( var n 5 ) " +
                                "( var r ( apply f x ) ) " +
                                "r ) ) ) " +
           "( apply g 2 ) )",  // result =  2 with static scope
                               // result = 10 with dynamic scope

         // recursive function
         "( prog " +
           "( fun fac ( lambda n " +
              "( if ( <= n 0 ) " +
                   "1 " +
                   "( * n ( apply fac ( - n 1 ) ) ) ) ) ) " +
           "( apply fac 6 ) )",

         // recursive function
         "( prog " +
           "( fun even ( lambda n " +
              "( && ( != n 1 ) " +       // use short-circuiting
                   "( || ( <= n 0 ) " +  // use short-circuiting
                        "( apply even ( - n 2 ) ) ) ) ) ) " +
           "( print ( apply even 7 ) ) " +
           "( print ( apply even 6 ) ) " +
           "( print ( apply even 0 ) ) " +
           "( apply even -1 ) )",

         // mutually recursive functions
         "( prog " +
           "( fun even ( lambda n " +
              "( || ( <= n 0 ) " +       // use short-circuiting
                   "( apply odd ( - n 1 ) ) ) ) ) " +
           "( fun odd ( lambda n " +
              "( && ( >  n 0 ) " +        // use short-circuiting
                   "( || ( == n 1 ) " +   // use short-circuiting
                        "( apply even ( - n 1 ) ) ) ) ) ) " +
           "( print ( apply even 8 ) ) " +
           "( print ( apply odd 7 ) ) " +
           "( print ( apply odd 6 ) ) " +
           "( print ( apply even 3 ) ) " +
           "( apply even -1 ) )",
      };


      int i = 0;
      for (i = 0; i < programs.length; i++)
      {
         System.out.println(i + " =============================================");

         // Print the expression.
         //System.out.println( programs[i] + "\n" );

         // Build the abstract syntax tree that represents the expression.
         try
         {
            Tree ast = BuildTree.buildTree( programs[i] );

            // Print the AST as an S-expression
            System.out.println( ast + "\n" );

            // Pretty-print the abstract syntax tree.
            System.out.println( PrettyPrinter.prettyPrint( ast ) + "\n" );

            // Print the infix version of the expression.
            System.out.println( AST2infix.ast2infix( ast ) + "\n" );

            // Evaluate the expression (interpret the AST).
            try
            {
               Value value = Evaluate.eval( ast );
               System.out.println( "result = " + value + "\n" );
            }
            catch (EvalException e)
            {
               System.out.println(e);
             //e.printStackTrace();
            }


            // Create dot and png files from the AST.
            if(true)
            try
            {
               // Create the (empty) dot file.
               String baseName = String.format("Language_7.%02d", i);
               java.io.PrintWriter out = new java.io.PrintWriter(
                                            new java.io.File(baseName + ".dot") );
               // Put dot commands into the dot file
               out.println( Tree2dot_ver1.tree2dot(ast) + "\n" );
               out.close();
               // Create a command line for running dot.exe.
               String[] cmd = {"dot.exe",
                               "-Tpng",
                               baseName + ".dot",
                               "-o",
                               baseName + ".png"};
               // Execute the command line.
               java.lang.Runtime.getRuntime().exec(cmd);
            }
            catch (Exception e)
            {
               System.out.println( e );
            }
         }
         catch (ParseException e)
         {
            System.out.println(e);
          //e.printStackTrace();
         }
      }
   }//main()

}//Language_7_Examples