/*
   This program parses and then evaluates strings from Language_7.

   These examples demonstrate some unusual ways in which we can use apply.
*/

public class Language_7_Examples_3
{
   public static void main(String[] args)
   {
      String[] programs =
      {
         "( prog " +
           "( fun f ( lambda x ( * x x ) ) ) " +
           "( var g f ) " +
           "( var r1 ( apply g 5 ) ) )",  // returns 25

         "( prog " +
           "( fun f ( lambda x y ( * x y ) ) ) " +
           "( fun g ( lambda f ) ) " +     // g has no parameters, always returns f
           "( apply ( apply g ) 5 6 ) )",  // returns 30

         "( prog " +
           "( fun f ( lambda x ( * x x ) ) ) " +
           "( fun g ( lambda x ( * x x x ) ) ) " +
           "( var b true ) " +
           "( var r1 ( apply ( if b f g ) 5 ) ) " +  // returns 25
           "( print r1 ) " +
           "( set b false ) " +
           "( var r2 ( apply ( if b f g ) 5 ) ) )",  // returns 125

         "( prog " +
           "( fun f ( lambda x ( * x x ) ) ) " +
           "( fun g ( lambda x ( * x x x ) ) ) " +
           "( fun h ( lambda x ( if x f g ) ) ) " +       // returns either f or g
           "( var r1 ( apply ( apply h  true ) 5 ) ) " +  // returns 25
           "( print r1 ) " +
           "( var r2 ( apply ( apply h false ) 5 ) ) )",  // returns 125

         "( prog " +
           "( var y 2 ) " +
           "( fun f ( lambda x ( * x y ) ) ) " +   // y is a non-local reference
           "( fun g ( lambda x ( * x x y ) ) ) " +
           "( fun h ( lambda x ( if x f g ) ) ) " + // returns either f or g
           "( begin " +
             "( var y 3 ) " +
             "( var r1 ( apply ( apply h  true ) 5 ) ) " +  //return 15 (dyn) or 10 (static)
             "( print r1 ) " +
             "( var r2 ( apply ( apply h false ) 5 ) ) ) )",//return 75 (dyn) or 50 (static)

         "( prog " +
           "( var x 0 ) " +
           "( apply x 1 ) )",  // error, x is not a function value

         "( apply ( && true ( == 2 ( + 1 1 ) ) ) 1 )", // error, not a function value

         "( apply ( fun g ( lambda x ( * x x ) ) ) 5 )", // error, fun is not yet an expression

         "( apply ( lambda x ( * x x ) ) 5 )" // error, lambda is not yet an expression
      };

      int i = 0;
      for (i = 0; i < programs.length; i++)
      {
         System.out.println(i + " =============================================");

         // Print the expression.
         //System.out.println( programs[i] + "\n" );

         // Build the abstract syntax tree that represents the expression.
         try
         {
            Tree ast = BuildTree.buildTree( programs[i] );

            // Print the AST as an S-expression
            //System.out.println( ast + "\n" );

            // Pretty-print the abstract syntax tree.
            System.out.println( PrettyPrinter.prettyPrint( ast ) + "\n" );

            // Print the infix version of the expression.
            System.out.println( AST2infix.ast2infix( ast ) + "\n" );

            // Evaluate the expression (interpret the AST).
            try
            {
               Value value = Evaluate.eval( ast );
               System.out.println( "result = " + value + "\n" );
            }
            catch (EvalException e)
            {
               System.out.println(e);
             //e.printStackTrace();
            }


            // Create dot and png files from the AST.
            if(true)
            try
            {
               // Create the (empty) dot file.
               String baseName = String.format("Language_7_3.%02d", i);
               java.io.PrintWriter out = new java.io.PrintWriter(
                                            new java.io.File(baseName + ".dot") );
               // Put dot commands into the dot file
               out.println( Tree2dot_ver1.tree2dot(ast) + "\n" );
               out.close();
               // Create a command line for running dot.exe.
               String[] cmd = {"dot.exe",
                               "-Tpng",
                               baseName + ".dot",
                               "-o",
                               baseName + ".png"};
               // Execute the command line.
               java.lang.Runtime.getRuntime().exec(cmd);
            }
            catch (Exception e)
            {
               System.out.println( e );
            }
         }
         catch (ParseException e)
         {
            System.out.println(e);
          //e.printStackTrace();
         }
      }
   }//main()

}//Language_7_Examples_3